#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <string>
#include <vector>

using std::string;
using std::vector;

/* Nul FD class */
static class nul_fd
{
	int fd;
public:
	nul_fd() : fd(open("/dev/null", O_WRONLY)) { }
	~nul_fd() { if(fd>=0)close(fd); }
	operator int () { return fd; }
} nul_fd;

static const string tempfile()
{
	char Buf[128];
	sprintf(Buf, "raropt__%d.rar", getpid());
	return Buf;	
};

static double similarity_test(const vector<string> &fn)
{
	const string tmpfn = tempfile();
	
	int p = vfork();
	
	if(!p)
	{
		dup2(nul_fd, 1);
		dup2(nul_fd, 2);
		
		vector<const char *> argv(10 + fn.size());
		argv[0] = "rar";
		argv[1] = "a";
		argv[2] = tmpfn.c_str();
		argv[3] = "-mm";
		argv[4] = "-m5";
		argv[5] = "-s";
		argv[6] = "-ol";
		argv[7] = "-ds";
		argv[8] = "--";
		argv[8 + fn.size()] = NULL;
		
		for(unsigned a=0; a<fn.size(); ++a)
			argv[8+a] = fn[a].c_str();
		
		execvp("rar", const_cast<char *const *> (&argv[0]));
	}

	struct stat st;
	unsigned total = 0;
	for(unsigned a=0; a<fn.size(); ++a)
	{
		stat(fn[a].c_str(), &st);
		total += st.st_size;
	}

	wait(NULL);
	
	stat(tmpfn.c_str(), &st);
	unlink(tmpfn.c_str());
	
	return total / (double)st.st_size;	
}

static double similarity_test(const string &fn1, const string &fn2)
{
	vector<string> fn(2);
	fn[0] = fn1;
	fn[1] = fn2;
	return similarity_test(fn);
}

static class ratios
{
	vector<string> files;
	vector<bool>   tested;
	vector<double> value;
	bool inited;
	
	void init()
	{
		tested.resize(files.size() * files.size(), false);
		value.resize(files.size() * files.size(), 0.0);
	}
	
public:
	ratios() : inited(false) { }
	
	void add_file(const string &fn)
	{
		files.push_back(fn);
	}
	
	const unsigned size() const { return files.size(); }
	
	const string &fn(unsigned id) const
	{
		return files[id];
	}
	
	const double get(unsigned a, unsigned b)
	{
		if(!inited) { init(); inited = true; }
		
		if(!tested[a*size() + b])
		{
			value[a*size() + b] = similarity_test(files[a], files[b]);
			tested[a*size() + b] = true;
		}
		return value[a*size() + b];
	}
} ratios;

int main(int argc, const char *const *argv)
{
	bool opts=true;
	bool worst=false;
	for(int a=1; a<argc; ++a)
	{
		const char *s = argv[a];
		if(opts && *s=='-' && s[1])
		{
			if(s[1]=='-') { if(!s[2])opts = false;
				else if(!strcmp(s+2, "--help")){s="h";goto Opt_H;}
				else if(!strcmp(s+2, "--version")){s="V";goto Opt_V;}
				else if(!strcmp(s+2, "--worst")){s="w";goto Opt_W;}
				else {
					fprintf(stderr, "Invalid switch: `%s'\n", s);
					return -1;
				}
			}
			else while(s[1])
				switch(*++s)
				{
					case 'w': Opt_W:
						worst = true;
						break;
					case 'v': Opt_V:
						printf("%s\n", VERSION);
						return 0;
					case 'h': Opt_H:
						printf(
							"Optrar v"VERSION" - Copyright (C) 1992,2001 Bisqwit (http://iki.fi/bisqwit/)\n"
							"\n"
							"Usage: optrar [-w | --worst] <file> [<...>]\n"
							"\n"
							"Builds the optimal list of files for solid rar archiving.\n"
							"\n"
							"These commands were used when testing:\n"
							"  rar a test.rar      -mm -m5 -s              main* *ak* optrar .depend\n"
							"     Result: 92061 bytes (default usage)\n"
							"  rar a worst.rar -ds -mm -m5 -s `./optrar -w main* *ak* optrar .depend`\n"
							"     Result: 93101 bytes (worst case)\n"
							"  rar a best.rar  -ds -mm -m5 -s `./optrar    main* *ak* optrar .depend`\n"
							"     Result: 92177 bytes (optimum case)\n"
							"\n"
							"Notice that all those rar-files contain the same files inside them,\n"
							"but the ordering of the files affects the size of the result.\n"
							"\n");
						return 0;
					default:
						fprintf(stderr, "Invalid option: `%c'\n", *s);
						return -1;
				}
		}
		else
		{
			ratios.add_file(s);
		}
	}
	
	const unsigned fc = ratios.size();
	
	vector< vector<string> > combination(fc);
	vector<double>           combscore(fc);
	
	// Find the best way to start
	for(unsigned first=0; first<fc; ++first)
	{
    	vector<bool> eliminated(fc, false);
    	unsigned prev = first;
    	for(unsigned elimicount=0; ; )
    	{
			fprintf(stderr, "\r%.2f%% done", (first*fc+elimicount)*100.0/(fc*fc));

    		combination[first].push_back(ratios.fn(prev).c_str());
    		if(++elimicount == fc)break;
    		eliminated[prev] = true;
    		
    		double best = 0;
    		if(worst)best = 3.4E+38;
    		unsigned bestp = 0;
    		for(unsigned a=0; a<fc; ++a)
    		{
    			if(eliminated[a])continue;
    			double ratio = ratios.get(prev, a);
    			if((!worst && ratio > best)
    			||  (worst && ratio < best))
    			{
    				bestp = a;
    				best = ratio;
    			}
    		}
    		prev = bestp;
    	}
    	
    	combscore[first] = similarity_test(combination[first]);
    	//fprintf(stderr, "Combination %u: score=%g\n", first, combscore[first]);
	}
	fprintf(stderr, "%-20s\r", "\r100% done\r"); fflush(stderr);
	
	double best = 0;
	if(worst)best = 3.4E+38;
	unsigned bestp = 0;
	for(unsigned a=0; a<fc; ++a)
	{
		if((!worst && combscore[a] > best)
		||  (worst && combscore[a] < best))
		{
			bestp = a;
			best = combscore[a];
		}
	}
	//fprintf(stderr, "Selected combination %u\n", bestp);
	
	for(unsigned a=0; a<fc; ++a)
		printf("%s\n", combination[bestp][a].c_str());

	fprintf(stderr, "\n");
	
	return 0;
}
